open Xstring
open Interfaces
open Definitions_abstract

class de_bruijn_terms _gmod =
  object (self)
    inherit abstract_terms_definition _gmod
	
	(** the definition title *)
    method title : string = "De Bruijn primitive syntax"
	(** the category of generated index case *)
    method index_type : string = "nat"
	
	(** generation of shifts *)
    method shift_terms_gen (type_name, level) : string = ""
end

class lift_fun _gmod =
  object (self)
    inherit abstract_fun_definition _gmod
	
	(** the definition title *)
    method title : string = "De Bruijn lifting function"
	(** function parameters list (name * type_name) *)
    method fun_params : (string * string) list =
      ("_n", "nat") :: ("_m", "nat") :: []
					 
					 (** return type of the function *)
    method return_type : string =
      self#get_target#get_name
	
	(** name of the function in a given target grammar *)
    method fun_name_type_name (name : string) : string =
      self#get_source#get_name ^ "_lift_in_" ^ name
						 
						 (** the source grammars for which generetion is needed.
						  * In this case : every indexed grammar. *)
    method get_source_grams : grammar option list =
      List.map (fun g -> Some g) gmod#get_indexed_grams
	
	(** the target (processed) grammars for which generetion is needed.
	 * By default, every grammar with the source grammar in its throughs. *)
    method get_target_grams (grams : grammar list) : grammar list =
      List.filter
	(fun g ->
	  (g = self#get_source) ||
	  (List.mem self#get_source g#get_throughs))
	grams
	
	(** generation of shifts *)
    method shift_fun_gen (gram, level) : string =
      if gram = self#get_source
      then level
      else ""
	  
	  (** generation of parameters *)
    method parameter_function_gen (p : parameter) : string =
      let shift =
	let raw_shifts = List.map self#shift_fun_gen p#get_shifts in
	let non_empty_shifts = List.filter (fun s -> s <> "") raw_shifts in
	String.concat " + " non_empty_shifts in
      let subst =
	("#param#", p#get_name) ::
	("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
	("#shift#", if shift = "" then "_m" else "(_m + " ^ shift ^ ")") ::
	[]
      and text =
	if self#is_self_rec p || self#is_back_rec p || self#is_throughs_rec p
	then
	  "(#recfun# _n #shift# #param#)"
	else
	  "#param#" in
      Xstring.replace subst text
	
	(** generation of constructors rhs : index case *)
    method index_constructor_rhs_function_gen (c : constructor) : string list =
      let subst =
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	[]
      and text =
	if self#get_target = self#get_source
	then
	  "if le_lt_dec _m #index# then #constr# (#index# + _n) else #constr# #index#" ::
	  []
	else "#constr# #index#" :: [] in
      Xstring.replaces subst [] text
	
	(** generation of index lemmas and tactics *)
    method index_lemmas_and_tactics (g : grammar) : (string list * string list) =
      let _ = self#set_source g in
      let _ = self#set_target g in
      let subst =
	("#lift#", self#fun_name) ::
	("#constr#",
	 match g#get_index_constructor_name with
	   Some s -> s
	 | None -> failwith ("definitions.lift_fun.index_lemmas_and_tactics : no index constructor in " ^ g#get_name)) ::
	[]
      and text =
	"  Lemma #lift#_#constr#_dec :" ::
	"    forall _n _m _p : nat," ::
	"    { prod (_m <= _p) (#lift# _n _m (#constr# _p) = #constr# (_p + _n)) } +" ::
	"    { prod (_p < _m) (#lift# _n _m (#constr# _p) = #constr# _p) }." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; subst; simpl; auto." ::
	"  Qed." ::
	"" ::
	"  Ltac _#lift#_#constr#_dec_tac n m p :=" ::
	"    let Hcomp := fresh \"Hcomp\" in" ::
	"          let Hfun := fresh \"Hfun\" in" ::
	"            destruct (#lift#_#constr#_dec n m p) as" ::
	"              [ [ Hcomp Hfun ] | [ Hcomp Hfun ] ];" ::
	"              rewrite ?Hfun in *; clear Hfun." ::
	[]
      and text2 =
	"      | [ |- context [ #lift# ?n ?m (#constr# ?p) ] ] =>" ::
	"        _#lift#_#constr#_dec_tac n m p" ::
	"" ::
	"      | [ H : context [ #lift# ?n ?m (#constr# ?p) ] |- _ ] =>" ::
	"        _#lift#_#constr#_dec_tac n m p" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
	
	(** generation of index equation lemmas *)
    method index_equation_lemmas (g : grammar) : (string list * string list) =
      let _ = self#set_source g in
      let _ = self#set_target g in
      let subst =
	("#lift#", self#fun_name) ::
	("#constr#",
	 match g#get_index_constructor_name with
	   Some s -> s
	 | None -> failwith ("definitions.lift_fun.index_lemmas_and_tactics : no index constructor in " ^ g#get_name)) ::
	[]
      and text =
	"  Lemma #lift#_#constr#_le :" ::
	"    forall _n _m _p : nat," ::
	"      _m <= _p -> #lift# _n _m (#constr# _p) = #constr# (_p + _n)." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; f_equal; omega." ::
	"  Qed." ::
	"" ::
	"  Lemma #lift#_#constr#_lt :" ::
	"    forall _n _m _p : nat," ::
	"      _p < _m -> #lift# _n _m (#constr# _p) = #constr# _p." ::
	"  Proof." ::
	"    intros; simpl; destruct le_lt_dec; f_equal; omega." ::
	"  Qed." ::
	[]
      and text2 =
	"      | [ |- context [ #lift# ?n ?m (#constr# ?p) ] ] =>" ::
	"          (cutrewrite ->" ::
	"            (#lift# n m (#constr# p) = #constr# (p + n));" ::
	"            [ | apply (#lift#_#constr#_le n m p); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#lift# n m (#constr# p) = #constr# p);" ::
	"            [ | apply (#lift#_#constr#_lt n m p); cfail ]) ||" ::
	"          lift_index_dec_tac" ::
	"" ::
	"      | [ H : context [ #lift# ?n ?m (#constr# ?p) ] |- _ ] =>" ::
	"          (cutrewrite ->" ::
	"            (#lift# n m (#constr# p) = #constr# (p + n)) in H;" ::
	"            [ | apply (#lift#_#constr#_le n m p); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#lift# n m (#constr# p) = #constr# p) in H;" ::
	"            [ | apply (#lift#_#constr#_lt n m p); cfail ]) ||" ::
	"          lift_index_dec_tac" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
  end

class lift_fun_with_source_and_target _gmod _source _target =
  object (self)
    inherit lift_fun _gmod as super
	
    initializer
      begin
	super#set_source _source;
	super#set_target _target
      end
  end

class subst_fun _gmod =
  object (self)
    inherit lift_fun _gmod
	
	(** the definition title *)
    method title : string = "De Bruijn substitution function"
	(** function parameters list (name * type_name) *)
    method fun_params : (string * string) list =
      ("_a", self#get_source#get_name) :: ("_n", "nat") :: []
							    
	(** name of the function in a given target grammar *)
    method fun_name_type_name (name : string) : string =
      self#get_source#get_name ^ "_subst_in_" ^ name
						  
	(** generation of parameters *)
    method private parameter_function_gen (p : parameter) : string =
      let shift =
	let raw_shifts = List.map self#shift_fun_gen p#get_shifts in
	let non_empty_shifts = List.filter (fun s -> s <> "") raw_shifts in
	String.concat " + " non_empty_shifts in
      let lift_rec =
	p#fold_shift
	  (fun cat -> fun level -> fun r ->
	    if self#get_source = cat
	  || List.mem cat self#get_source#get_throughs then
	      let lift =
		new lift_fun_with_source_and_target gmod cat self#get_source in
	      "(" ^ lift#fun_name ^ " " ^ level ^ " 0 " ^ r ^ ")"
	    else r)
	  "_a" in
      let subst =
	("#param#", p#get_name) ::
	("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
	("#shift#", if shift = "" then "_n" else "(_n + " ^ shift ^ ")") ::
	("#lift_rec#", lift_rec) ::
	[]
      and text =
	if self#is_self_rec p || self#is_back_rec p || self#is_throughs_rec p
	then
	  "(#recfun# #lift_rec# #shift# #param#)"
	else
	  "#param#" in
      Xstring.replace subst text
	
	(** generation of constructors rhs : index case *)
    method private index_constructor_rhs_function_gen (c : constructor) : string list =
      let subst =
	("#constr#", c#get_name) ::
	("#index#", c#get_index_name) ::
	[]
      and text =
	if self#get_target = self#get_source
	then
	  "match lt_eq_lt_dec _n #index# with" ::
	  "    (* case _n < #index# *)" ::
	  "  | inleft (left _) => #constr# (#index# - 1)" ::
	  "    (* case _n = #index# *)" ::
	  "  | inleft (right _) => _a" ::
	  "    (* case #index# < _n *)" ::
	  "  | inright _ => #constr# #index#" ::
	  "end" ::
	  []
	else "#constr# #index#" :: [] in
      Xstring.replaces subst [] text
	
	(** generation of index lemmas and tactics *)
    method index_lemmas_and_tactics (g : grammar) : (string list * string list) =
      let _ = self#set_source g in
      let _ = self#set_target g in
      let subst =
	("#subst#", self#fun_name) ::
	("#type#", g#get_name) ::
	("#constr#",
	 match g#get_index_constructor_name with
	   Some s -> s
	 | None -> failwith ("definitions.subst_fun.index_lemmas_and_tactics : no index constructor in " ^ g#get_name)) ::
	[]
      and text =
	"  Lemma #subst#_#constr#_dec :" ::
	"    forall (_a : #type#) (_n _p : nat)," ::
	"    { prod (_n < _p) (#subst# _a _n (#constr# _p) = #constr# (_p - 1)) } +" ::
	"    { prod (_n = _p) (#subst# _a _n (#constr# _p) = _a) } +" ::
	"    { prod (_p < _n) (#subst# _a _n (#constr# _p) = #constr# _p) }." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ];" ::
	"      subst; simpl; #crush#." ::
	"  Qed." ::
	"" ::
	"  Ltac _#subst#_#constr#_dec_tac a n p :=" ::
	"    let Hcomp := fresh \"Hcomp\" in" ::
	"          let Hfun := fresh \"Hfun\" in" ::
	"            destruct (#subst#_#constr#_dec a n p) as" ::
	"              [ [ [ Hcomp Hfun ] | [ Hcomp Hfun ] ]" ::
	"                | [ Hcomp Hfun ] ];" ::
	"              rewrite ?Hfun in *; clear Hfun." ::
	[]
      and text2 =
	"      | [ |- context [ #subst# ?a ?n (#constr# ?p) ] ] =>" ::
	"        _#subst#_#constr#_dec_tac a n p" ::
	"" ::
	"      | [ H : context [ #subst# ?a ?n (#constr# ?p) ] |- _ ] =>" ::
	"        _#subst#_#constr#_dec_tac a n p" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
	
	(** generation of index equation lemmas *)
    method index_equation_lemmas (g : grammar) : (string list * string list) =
      let _ = self#set_source g in
      let _ = self#set_target g in
      let subst =
	("#subst#", self#fun_name) ::
	("#type#", g#get_name) ::
	("#constr#",
	 match g#get_index_constructor_name with
	   Some s -> s
	 | None -> failwith ("definitions.lift_fun.index_lemmas_and_tactics : no index constructor in " ^ g#get_name)) ::
	[]
      and text =
	"  Lemma #subst#_#constr#_lt_l :" ::
	"    forall (_a : #type#) (_n _p : nat)," ::
	"      _n < _p -> #subst# _a _n (#constr# _p) = #constr# (_p - 1)." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ]; #crush#." ::
	"  Qed." ::
	"" ::
	"  Lemma #subst#_#constr#_eq :" ::
	"    forall (_a : #type#) (_n _p : nat)," ::
	"      _n = _p -> #subst# _a _n (#constr# _p) = _a." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ]; #crush#." ::
	"  Qed." ::
	"" ::
	"  Lemma #subst#_#constr#_lt_r :" ::
	"    forall (_a : #type#) (_n _p : nat)," ::
	"      _p < _n -> #subst# _a _n (#constr# _p) = #constr# _p." ::
	"  Proof." ::
	"    intros; simpl; destruct lt_eq_lt_dec as [ [ ? | ? ] | ? ]; #crush#." ::
	"  Qed." ::
	[]
      and text2 =
	"        | [ |- context [ #subst# ?a ?n (#constr# ?m) ] ] =>" ::
	"          (cutrewrite ->" ::
	"            (#subst# a n (#constr# m) = #constr# (m - 1));" ::
	"            [ | apply (#subst#_#constr#_lt_l a n m); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# a n (#constr# m) = a);" ::
	"            [ | apply (#subst#_#constr#_eq a n m); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# a n (#constr# m) = #constr# m);" ::
	"            [ | apply (#subst#_#constr#_lt_r a n m); cfail ]) ||" ::
	"          subst_index_dec_tac" ::
	"" ::
	"        | [ H : context [ #subst# ?a ?n (#constr# ?m) ] |- _ ] =>" ::
	"          (cutrewrite ->" ::
	"            (#subst# a n (#constr# m) = #constr# (m - 1)) in H;" ::
	"            [ | apply (#subst#_#constr#_lt_l a n m); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# a n (#constr# m) = a) in H;" ::
	"            [ | apply (#subst#_#constr#_eq a n m); cfail ]) ||" ::
	"          (cutrewrite ->" ::
	"            (#subst# a n (#constr# m) = #constr# m) in H;" ::
	"            [ | apply (#subst#_#constr#_lt_r a n m); cfail ]) ||" ::
	"          subst_index_dec_tac" ::
	"" ::
	[] in
      (Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
  end

class subst_fun_with_source_and_target _gmod _source _target =
object
	inherit subst_fun _gmod as super
	
	initializer
	begin
		super#set_source _source;
		super#set_target _target
	end
end
