open Xlist
open Xstring

open Interfaces

(** class of grammars *)
class grammar_impl _gmod _name _constrs : grammar =
  object (self)
      (** grammar module of this element *)
    val mutable gmod : gmodule = _gmod
	
	(** grammar (category) name *)
    val name : string = _name
	(** constructors of this grammar *)
    val constrs : constructor list = _constrs
	(** attaignable categories with indices names *)
    val mutable throughs : grammar list = []
	
	(** accessors *)
    method get_name : string = name
    method get_throughs : grammar list = throughs
    method get_constrs : constructor list = constrs
	
	(** post-initialization with the given gmodule *)
    method set_gmodule (gm : gmodule) : unit =
      begin
	(* set the gmod field *)
	gmod <- gm;
	(* performs post_initialization for constructors *)
	List.iter (fun c -> c#set_gmodule gm) constrs
      end
	
	(** collects the name of the De Bruijn variable constructor in a grammar *)
    method get_index_constructor_name : string option =
      let rec aux = fun (cstrs : constructor list) ->
	match cstrs with
	  [] -> None
	| c :: cs -> if c#is_index then Some c#get_name else aux cs
      in aux constrs
	
	(** return true/false whether the grammar contains De Bruijn variables *)
    method has_index : bool =
      match self#get_index_constructor_name with
	Some _ -> true
      | None -> false
	    
    method private get_sub_throughs : grammar list =
      let all_throughs =
	List.fold_right (fun c -> fun l -> c#get_throughs @ l) constrs []
      in
      Xlist.remove (self :> grammar) (Xlist.doublon_remove all_throughs)
	
	(** initializes the through list, collecting immediate
	 * subcategories that has index *)
    method throughs_init : unit =
      throughs <- self#get_sub_throughs
	  
	  (** expands the through list, adding immediate
	   * subcategories throughs, returns true if a modification
	   * is performed *)
    method throughs_expand : bool =
      let new_throughs = self#get_sub_throughs in
      if Xlist.multiset_equal throughs new_throughs then false
      else
	begin
	  throughs <- Xlist.doublon_remove (throughs @ new_throughs);
	  true
	end
	  
	  (** induction patterns for this grammar *)
    method private induction_patterns : string list list =
      List.map (fun c -> c#induction_pattern (self :> grammar)) constrs
	
	(** induction principle invocation for this grammar for a given argument *)
    method by_induction (arg : string) : string =
      let patterns =
	String.concat " | "
	  (List.map (String.concat " ") self#induction_patterns) in
      "induction " ^ arg ^ " as [ " ^ patterns ^ " ]"
						   
						   (** rewrite induction hypothesis for this grammar *)
    method rewrite_induction (prop : property) : string list =
      let constrs_ind =
	List.map (fun c -> c#rewrite_induction prop (self :> grammar)) constrs in
      let constrs_ind_with_crush =
	List.map (fun s -> if s = "" then "#crush#" else s ^ "; #crush#")
	  constrs_ind in
      let rew_hd = List.hd constrs_ind_with_crush
      and rew_tl =
	List.map (fun s -> "        | " ^ s)
	  (List.tl constrs_ind_with_crush) in
      ("        [ " ^ rew_hd) ::
      rew_tl @
      "        ];" ::
      []
	
	(** function equations generation *)
    method fun_equations_gen (def : fun_definition) : string list =
      let _ = def#set_target (self :> grammar) in
      ("  (** Equations for " ^ def#title ^ " *)") :: "" ::
      (Xlist.concat_with [""] (List.map (fun c -> c#fun_equation_gen def) constrs))
	
	(** property generation *)
    method property_gen (prop : property) (prefix : string) (n : node)
	: string list * string list * string list * (string * (string list * string list)) =
      let _ = prop#set_node n in
      let _ = prop#set_target (self :> grammar) in
      let subst =
	("#Lemma#", prefix) ::
	("#lemma_name#", prop#lemma_name) ::
	("#target_type#", self#get_name) ::
	("#source_type#", if prop#has_source then prop#get_source#get_name else "") ::
	("#index_constr#",
	 if prop#has_source then
	   match prop#get_source#get_index_constructor_name with
	     Some name -> name
	   | None -> ""
	 else "") ::
	("#by_induction#", self#by_induction "_arg") ::
	prop#subst
      and substs =
	("#rewrite_induction_hypothesis#",
	 self#rewrite_induction prop) ::
	[] in
      let lemma_header = Xstring.replaces subst substs prop#lemma_header
      and lemma_proof =
	let proof_begin =
	  if List.length (prop#get_target_grams prop#get_node#get_grams) = 1
	  then []
	  else
	    "    (* case #lemma_name# *)" ::
	    "    Focus. clear #lemma_name#." ::
	    [] in
	Xstring.replaces subst substs (proof_begin @ prop#lemma_proof)
      and aux_defs = Xstring.replaces subst substs prop#aux_defs
      and tactic_template =
	let (pat, (body, post)) = prop#tactic_template in
	(Xstring.replace subst pat,
	 (Xstring.replaces subst substs body,
	  Xstring.replaces subst substs post)) in
      (lemma_header, lemma_proof, aux_defs, tactic_template)
	
	(** Pretty printing *)
    method to_string () : string =
      name ^
      " [ has_index : " ^ (string_of_bool self#has_index) ^ " ; throughs : " ^
      (String.concat ", " (List.map (fun g -> g#get_name) throughs)) ^ " ]" ^
      " : Set :=\n" ^
      (String.concat "\n" (List.map (fun c -> c#to_string ()) constrs))
  end

(** class of external types *)
class external_impl _name : grammar =
object (self)
	(** grammar (category) name *)
	val name : string = _name
	
	method get_name : string = name
	method get_throughs : grammar list = []
	method get_constrs : constructor list = []
	method get_index_constructor_name : string option = None
	method has_index : bool = false
	
	method set_gmodule : gmodule -> unit =
		failwith "Grammar.external_impl: illicit method invocation"
	method throughs_init : unit =
		failwith "Grammar.external_impl: illicit method invocation"
	method throughs_expand : bool =
		failwith "Grammar.external_impl: illicit method invocation"
	method by_induction (arg : string) : string =
		failwith "Grammar.external_impl: illicit method invocation"
	method rewrite_induction (prop : property) : string list =
		failwith "Grammar.external_impl: illicit method invocation"
	method fun_equations_gen (def : fun_definition) : string list =
		 failwith "Grammar.external_impl: illicit method invocation"
	method property_gen (prop : property) (prefix : string) (n : node)
	: string list * string list * string list * (string * (string list * string list)) =
		 failwith "Grammar.external_impl: illicit method invocation"
			
	(** Pretty printing *)
	method to_string () : string =
		name
end
