open Debug
open Xlist

open Interfaces

(** class of constructor parameters *)
class parameter_impl _gmod _shifts _name _gtype : parameter =
  object
      (** grammar module of this element *)
    val mutable gmod : gmodule = _gmod
	
	(** binding categories (type_name, shift_level) *)
    val mutable shifts : (grammar * string) list =
      Xlist.assoc_aggregate (String.concat " + ") _shifts
	(** parameter name used for pretty printing *)
    val name : string = _name
	(** category name of the parameter *)
    val mutable gtype : grammar = _gtype
	
	(** accessors *)
    method get_name : string = name
    method get_type : grammar = gtype
    method get_shifts : (grammar * string) list = shifts
	
	(** returns true/false whether this is an index parameter *)
    method is_index : bool = false
	
	(** post-initialization with the given gmodule *)
    method set_gmodule (gm : gmodule) : unit =
      (* returns the gmodule grammar replacing this grammar, if any *)
      let get_gmodule_gram g =
	if gm#is_gmodule_gram g#get_name
	then gm#get_gram g#get_name
	else g in
      begin
	(* set the gmod field *)
	gmod <- gm;
	(* connect gtype and shifted grammars *)
	gtype <- get_gmodule_gram gtype;
	shifts <- List.map (fun (g, n) -> (get_gmodule_gram g, n)) shifts
      end
	
	(** returns the through list of the category, i.e.
	 * the category name is it has index, and the throughs list
	 * of the category *)
    method get_throughs : grammar list =
      (if gtype#has_index then [gtype] else []) @
      gtype#get_throughs
	
	(** induction pattern for this parameter *)
    method induction_pattern (g : grammar) : string list =
      name :: if gtype = g then [ "IH" ^ name ] else []
	
	(** rewrite induction hypothesis for this parameter *)
    method rewrite_induction (prop : property) (source : grammar) : string =
      if gtype = source
      then "simpl; try rewrite IH" ^ name
      else
	if List.mem gtype (prop#get_target_grams gmod#get_all_grams)
	then "simpl; try rewrite " ^ prop#lemma_name_type_name gtype#get_name
	else "simpl"
	
	(** fold a function on the shift list *)
    method fold_shift (f : grammar -> string -> string -> string) (init : string) : string =
      List.fold_right (fun (cat, level) -> fun r -> f cat level r) shifts init
	    
	    (** Pretty printing *)
    method to_string () : string =
      " (" ^
      (if List.length shifts = 0 then "" else
      "(* bind " ^
      (String.concat ", " (List.map (fun (g, n) -> n ^ " " ^ g#get_name) shifts)) ^
      " in *) ") ^
      name ^ " : " ^ gtype#get_name ^
      ")"
  end

(** class of index constructor parameters *)
class index_parameter_impl _gmod _name _gtype : parameter =
  object
    inherit parameter_impl _gmod [] _name _gtype
	
	(** returns true/false whether this is an index parameter *)
    method is_index : bool = true
	
	(** rewrite induction hypothesis for this parameter *)
    method rewrite_induction (prop : property) (g : grammar) : string =
      prop#get_index_tac
	
	(** Pretty printing *)
    method to_string () : string =
      "((* index *) " ^ name ^ " : " ^ gtype#get_name ^ ")"
  end
